/*
 * ni_virt.c - Virtualization monitor plugin
 *
 *  (c)2008,2010 Todd Shadburn
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#include <sys/time.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <pthread.h>

#include <cutil/slist.h>

#include <libvirt/libvirt.h>

#include <netinfo/netinfo.h>
#include <netinfo/nmodule.h>
#include <netinfo/conffile.h>
#include <netinfo/event.h>


#define	MODULE_NAME			"virt"
#define	MODULE_VERSION			"0.1.1"

#define	PASSIVE_CONFIG_FILE		"passive.conf"

#define	EVENTRC_OK			0
#define	EVENTRC_WARNING			1
#define	EVENTRC_CRITICAL		2


#define	VIRT_DOMAIN_OBJECT		"DOMAIN"

#define	MAX_DOMAINS			256
#define	DEFAULT_MAX_DOMAIN_TIMEDIFF	60
#define	DEFAULT_POLL_INTERVAL		60

#define	VIRTMOD_FLAG_EXIT_SIGNALED	1

#define	BUFFER_SIZE			256


typedef struct	_DomainData		DomainData;
struct	_DomainData {
	int		id;
	time_t		timestamp;
	char		*name;
	char		uuid[VIR_UUID_STRING_BUFLEN];
	virDomainInfo	info;
};


struct {
	ModuleInfo	*moduleinfo;
	EventChain	*eventchain;
	EventData	*event;

	int		pollinterval;
	pthread_t	polltid;

	pthread_mutex_t	mutex;
	pthread_cond_t	cond;
	uint32_t	flags;
	time_t		max_domain_timediff;

	virConnectPtr	conn;
	SList		*domainlist;

	char		msg_buffer[BUFFER_SIZE];
	int		msg_len;

} VirtModule;



DomainData *module_lookup_domain_by_id (int id)
{
	DomainData	*dd = NULL;
	SList		*litem = VirtModule.domainlist;
	while (litem != NULL) {
		dd = slist_get_data (litem);
		if (dd->id == id)
			return dd;
		litem = slist_get_next (litem);
	}
	return NULL;
}


DomainData *module_lookup_domain_by_name (char *name)
{
	DomainData	*dd = NULL;
	SList		*litem = VirtModule.domainlist;
	while (litem != NULL) {
		dd = slist_get_data (litem);
		if (!strcmp (dd->name, name))
			return dd;
		litem = slist_get_next (litem);
	}
	return NULL;
}


int module_add_domain (int domid)
{
	int		rc;
	DomainData	*dd = NULL;
	virDomainPtr	dom = NULL;

	dom = virDomainLookupByID (VirtModule.conn, domid);
	if (dom == NULL)
		return EINVAL;

	dd = calloc (1, sizeof (DomainData));
	if (dd == NULL)
		return ENOMEM;

	dd->id = domid;
	dd->name = strdup (virDomainGetName (dom));
	rc = virDomainGetUUIDString (dom, dd->uuid);
	rc = virDomainGetInfo (dom, &dd->info);
	dd->timestamp = time (NULL);

	rc = slist_item_add (&VirtModule.domainlist, dd);
	vector_log_entry (VirtModule.moduleinfo,
			"Added domain '%s'\n", dd->name);
	return 0;
}


int module_remove_domain (int domid)
{
	int		rc;
	DomainData	*dd = NULL;
	dd = module_lookup_domain_by_id (domid);
	if (dd == NULL)
		return EINVAL;

	rc = slist_item_remove (&VirtModule.domainlist, dd);
	if (rc != 0)
		return rc;
	vector_log_entry (VirtModule.moduleinfo,
			"Removed domain '%s'\n", dd->name);

	if (dd->name)
		free (dd->name);
	free (dd);
	return 0;
}


int module_check_domain (int domid)
{
	int		rc;
	time_t		ct;
	uint32_t	eventrc;
	virDomainInfo	info;
	char		*name = NULL;
	DomainData	*dd = NULL;
	virDomainPtr	dom = NULL;

	ct = time (NULL);

	eventrc = EVENTRC_OK;
	dom = virDomainLookupByID (VirtModule.conn, domid);
	name = virDomainGetName (dom);
	rc = virDomainGetInfo (dom, &info);
	
	/* existance checks */
	dd = module_lookup_domain_by_id (domid);
	if (dd == NULL) {
		dd = module_lookup_domain_by_name (name);
		if (dd == NULL) {
			/*
			 * Newly booted domain
			*/
			rc = module_add_domain (domid);
			VirtModule.msg_len = snprintf (VirtModule.msg_buffer,
					BUFFER_SIZE,
					"%s domain is running",
					name);
			eventdata_clear (VirtModule.event);
			rc = eventdata_set (VirtModule.event,
					1,
					EVENTRC_OK,
					VIRT_DOMAIN_OBJECT,
					VirtModule.msg_buffer,
					VirtModule.msg_len);
			rc = vector_eventchain_emit_event (VirtModule.eventchain,
						VirtModule.event);
			return 0;
		} else {
			/*
			 * domain id changed (domain restart)
			*/
			dd->id = domid;
			VirtModule.msg_len = snprintf (VirtModule.msg_buffer,
					BUFFER_SIZE,
					"%s domain restarted",
					name);
			eventdata_clear (VirtModule.event);
			rc = eventdata_set (VirtModule.event,
					1,
					EVENTRC_OK,
					VIRT_DOMAIN_OBJECT,
					VirtModule.msg_buffer,
					VirtModule.msg_len);
			rc = vector_eventchain_emit_event (VirtModule.eventchain,
						VirtModule.event);
		}
	}

	/* movement check */
	if (dd->info.cpuTime == info.cpuTime) {
		/*
		 * stalled (or paused) domain
		*/
		if (info.state == VIR_DOMAIN_PAUSED) {
			VirtModule.msg_len = snprintf (VirtModule.msg_buffer,
				BUFFER_SIZE,
				"%s domain is paused",
				name);
			eventrc = EVENTRC_WARNING;
		} else {
			VirtModule.msg_len = snprintf (VirtModule.msg_buffer,
				BUFFER_SIZE,
				"%s domain is stalled",
				name);
			eventrc = EVENTRC_CRITICAL;
		}
		eventdata_clear (VirtModule.event);
		rc = eventdata_set (VirtModule.event,
				1,
				eventrc,
				VIRT_DOMAIN_OBJECT,
				VirtModule.msg_buffer,
				VirtModule.msg_len);
		rc = vector_eventchain_emit_event (VirtModule.eventchain,
					VirtModule.event);
	}

	/* post processing updates */
	memcpy (&dd->info, &info, sizeof (virDomainInfo));
	dd->timestamp = time (NULL);

	return 0;
}


int module_process_shutdown_domains (void)
{
	int		rc;
	int		i;
	int		dcount;
	time_t		ct;
	int		rdd[MAX_DOMAINS];
	DomainData	*dd = NULL;
	SList		*litem = NULL;

	ct = time (NULL);

	dcount = 0;
	litem = VirtModule.domainlist;
	while (litem) {
		dd = slist_get_data (litem);
		if ((ct - dd->timestamp) > VirtModule.max_domain_timediff) {
			/*
			 * domain is no longer running
			*/
			rdd[dcount] = dd->id;
			dcount++;
			VirtModule.msg_len = snprintf (VirtModule.msg_buffer,
					BUFFER_SIZE,
					"%s domain is no longer running",
					dd->name);
			eventdata_clear (VirtModule.event);
			rc = eventdata_set (VirtModule.event,
					1,
					EVENTRC_CRITICAL,
					VIRT_DOMAIN_OBJECT,
					VirtModule.msg_buffer,
					VirtModule.msg_len);
			rc = vector_eventchain_emit_event (VirtModule.eventchain,
						VirtModule.event);
		}
		litem = slist_get_next (litem);
	}
	for (i=0 ; i<dcount ; i++) {
		rc = module_remove_domain(rdd[i]);
	}


	return 0;
}


void *module_monitor_thread (void *arg)
{
	int		rc;
	int		i;
	int		count;
	int		domid[MAX_DOMAINS];
	struct timespec ts;
	struct timeval  tv;

	/* setup */
	rc = vector_eventchain_open (VirtModule.moduleinfo,
				&VirtModule.eventchain);

	/* Add running domains to the list */
	count = virConnectListDomains (VirtModule.conn,
				domid,
				MAX_DOMAINS);
	for (i=0 ; i<count ; i++) {
		rc = module_add_domain (domid[i]);
	}

	/*loop */
	while (1) {

		/* wait for timeout or cond */
		rc = gettimeofday (&tv, NULL);
		ts.tv_sec = tv.tv_sec + VirtModule.pollinterval;
		ts.tv_nsec = tv.tv_usec * 1000;
		pthread_mutex_lock (&VirtModule.mutex);
		pthread_cond_timedwait (&VirtModule.cond,
					&VirtModule.mutex,
					&ts);
		if (VirtModule.flags & VIRTMOD_FLAG_EXIT_SIGNALED) {
			pthread_mutex_unlock (&VirtModule.mutex);
			break;
		}
		pthread_mutex_unlock (&VirtModule.mutex);


		/* process running domains */
		count = virConnectListDomains (VirtModule.conn,
					domid,
					MAX_DOMAINS);
		for (i=0 ; i<count ; i++) {
			rc = module_check_domain (domid[i]);
		}

		/* process missing domains */
		rc = module_process_shutdown_domains ();

	}

	/* teardown */
	vector_eventchain_close (&VirtModule.eventchain);
	pthread_exit (NULL);
	return NULL;
}


int netinfo_module_init (ModuleInfo *mi)
{
	int		rc;

	memset (&VirtModule, 0, sizeof (VirtModule));
	VirtModule.moduleinfo = mi;
	VirtModule.moduleinfo->name = MODULE_NAME;

	/* set values or defaults */
	VirtModule.max_domain_timediff = DEFAULT_MAX_DOMAIN_TIMEDIFF;
	VirtModule.pollinterval = DEFAULT_POLL_INTERVAL;

	/* init structure elements */
	rc = pthread_mutex_init (&VirtModule.mutex, NULL);
	if (rc != 0)
		return NETINFO_RC_MODULEERROR;
	pthread_cond_init (&VirtModule.cond, NULL);
	if (rc != 0) {
		pthread_mutex_destroy (&VirtModule.mutex);
		return NETINFO_RC_MODULEERROR;
	}

	/* Make connection */
	VirtModule.conn = virConnectOpenReadOnly (NULL);
	if (VirtModule.conn == NULL) {
		pthread_mutex_destroy (&VirtModule.mutex);
		pthread_cond_destroy (&VirtModule.cond);
		return 1;
	}

	rc = eventdata_new (&VirtModule.event);
	if (rc != NETINFO_RC_SUCCESS) {
		pthread_mutex_destroy (&VirtModule.mutex);
		pthread_cond_destroy (&VirtModule.cond);
		return rc;
	}

	/* start thread */
	rc = pthread_create (&VirtModule.polltid,
				NULL,
				module_monitor_thread,
				NULL);
	if (rc == 0) {
		vector_log_entry (mi,
				"Started poll thread (tid=%lu) interval=%d\n",
				VirtModule.polltid,
				VirtModule.pollinterval);
	} else {
		vector_log_entry (mi,
				"Failed to start poll thread.\n");
		pthread_mutex_destroy (&VirtModule.mutex);
		pthread_cond_destroy (&VirtModule.cond);
		eventdata_free (&VirtModule.event);
		return NETINFO_RC_MODULEERROR;
	}

	return NETINFO_RC_SUCCESS;
}


int netinfo_module_close (ModuleInfo *mi)
{
	int		rc;
	int		i;
	int		count;
	int		domid[MAX_DOMAINS];
	SList		*litem = NULL;
	DomainData	*dd = NULL;

	/* signal the thread to exit */
	pthread_mutex_lock (&VirtModule.mutex);
	VirtModule.flags |= VIRTMOD_FLAG_EXIT_SIGNALED;
	pthread_mutex_unlock (&VirtModule.mutex);
	pthread_cond_signal (&VirtModule.cond);

	vector_log_entry (mi,
			"Shutting down poll thread.\n");

	/* free resources */
	count = 0;
	litem = VirtModule.domainlist;
	while (litem) {
		dd = slist_get_data (litem);
		domid[count] = dd->id;
		count++;
		litem = slist_get_next (litem);
	}
	for (i=0 ; i<count ; i++) {
		rc = module_remove_domain (domid[i]);
	}

	/* close connection */
	rc = virConnectClose (VirtModule.conn);

	return 0;
}



